//! This library provides a simple procedural macro for fast implementing serialize methods in
//! [serde::Serializer](https://docs.rs/serde/latest/serde/trait.Serializer.html) trait.
/// Implements serialize methods for [serde::Serializer](https://docs.rs/serde/latest/serde/trait.Serializer.html)
/// 
/// # Examples
///
/// ```
/// use impl_serialize::impl_serialize;
/// use serde::ser;
/// # use thiserror::Error;
/// # 
/// # #[derive(Debug, Error)]
/// # enum SerializationError {
/// #     #[error("Other error")]
/// #     OtherError,
/// #     #[error("Cannot serialize value from {0}")]
/// #     CannotSerializeFrom(String),
/// #     #[error("Custom({0})")]
/// #     Custom(String)
/// # }
/// # 
/// # impl serde::ser::Error for SerializationError {
/// #     fn custom<T>(msg:T) -> Self
/// #     where T: std::fmt::Display
/// #     {
/// #         SerializationError::Custom(msg.to_string())
/// #     }
/// # }
/// # 
/// struct MySerializer;
/// 
/// impl ser::Serializer for MySerializer {
/// #    type Ok = ();
/// #    type Error = SerializationError;
/// #
/// #   type SerializeMap = ser::Impossible<Self::Ok, Self::Error>;
/// #   type SerializeSeq = ser::Impossible<Self::Ok, Self::Error>;
/// #   type SerializeStruct = ser::Impossible<Self::Ok, Self::Error>;
/// #   type SerializeStructVariant = ser::Impossible<Self::Ok, Self::Error>;
/// #   type SerializeTuple = ser::Impossible<Self::Ok, Self::Error>;
/// #   type SerializeTupleStruct = ser::Impossible<Self::Ok, Self::Error>;
/// #   type SerializeTupleVariant = ser::Impossible<Self::Ok, Self::Error>;
/// #
///     //with value type
///     impl_serialize!(
///         |value_type: &str| {
///             Err(SerializationError::CannotSerializeFrom(value_type.to_string()))
///         },
///         bool
///     );
///     
///     //without value type
///     impl_serialize!(
///         |_| Err(SerializationError::OtherError),
///         char
///     );
/// 
///     //for many types
///     impl_serialize!(
///         |value_type: &str| {
///             Err(SerializationError::CannotSerializeFrom(value_type.to_string()))
///         },
///         [
///             bytes,
///             i8, i16, i32, i64,
///             // other types
/// #           u8, u16, u32, u64,
/// #           f32, f64,
/// #           str,
/// #           none, some, unit,
/// #           unit_struct, unit_variant,
/// #           newtype_struct, newtype_variant,
/// #           seq, map,
/// #           tuple, tuple_struct, tuple_variant,
/// #           struct, struct_variant
///         ]
///     );
/// }
/// ```

#[macro_export]
macro_rules! impl_serialize {
    ($get_result:expr, [$($type:ident),+]) => {
        $(
            impl_serialize!($get_result, $type);
        )+
    };

    ($get_result:expr, bool) => {
        fn serialize_bool(self, _v: bool) -> Result<Self::Ok, Self::Error> {
            $get_result("bool")
        }
    };

    ($get_result:expr, i8) => {
        fn serialize_i8(self, _v: i8) -> Result<Self::Ok, Self::Error> {
            $get_result("i8")
        }
    };

    ($get_result:expr, i16) => {
        fn serialize_i16(self, _v: i16) -> Result<Self::Ok, Self::Error> {
            $get_result("i16")
        }
    };

    ($get_result:expr, i32) => {
        fn serialize_i32(self, _v: i32) -> Result<Self::Ok, Self::Error> {
            $get_result("i32")
        }
    };

    ($get_result:expr, i64) => {
        fn serialize_i64(self, _v: i64) -> Result<Self::Ok, Self::Error> {
            $get_result("i64")
        }
    };

    ($get_result:expr, u8) => {
        fn serialize_u8(self, _v: u8) -> Result<Self::Ok, Self::Error> {
            $get_result("u8")
        }
    };

    ($get_result:expr, u16) => {
        fn serialize_u16(self, _v: u16) -> Result<Self::Ok, Self::Error> {
            $get_result("u16")
        }
    };

    ($get_result:expr, u32) => {
        fn serialize_u32(self, _v: u32) -> Result<Self::Ok, Self::Error> {
            $get_result("u32")
        }
    };

    ($get_result:expr, u64) => {
        fn serialize_u64(self, _v: u64) -> Result<Self::Ok, Self::Error> {
            $get_result("u64")
        }
    };

    ($get_result:expr, f32) => {
        fn serialize_f32(self, _v: f32) -> Result<Self::Ok, Self::Error> {
            $get_result("f32")
        }
    };

    ($get_result:expr, f64) => {
        fn serialize_f64(self, _v: f64) -> Result<Self::Ok, Self::Error> {
            $get_result("64")
        }
    };

    ($get_result:expr, char) => {
        fn serialize_char(self, _v: char) -> Result<Self::Ok, Self::Error> {
            $get_result("char")
        }
    };

    ($get_result:expr, str) => {
        fn serialize_str(self, _v: &str) -> Result<Self::Ok, Self::Error> {
            $get_result("str")
        }
    };

    ($get_result:expr, bytes) => {
        fn serialize_bytes(self, _v: &[u8]) -> Result<Self::Ok, Self::Error> {
            $get_result("bytes")
        }
    };

    ($get_result:expr, none) => {
        fn serialize_none(self) -> Result<Self::Ok, Self::Error> {
            $get_result("none")
        }
    };

    ($get_result:expr, unit) => {
        fn serialize_unit(self) -> Result<Self::Ok, Self::Error> {
            $get_result("unit")
        }
    };

    ($get_result:expr, some) => {
        fn serialize_some<T: ?Sized + serde::ser::Serialize>(self, _value: &T) -> Result<Self::Ok, Self::Error> {
            $get_result("some")
        }
    };

    ($get_result:expr, unit_struct) => {
        fn serialize_unit_struct(self, _name: &'static str) -> Result<Self::Ok, Self::Error> {
            $get_result("unit_struct")
        }
    };

    ($get_result:expr, unit_variant) => {
        fn serialize_unit_variant(self, _name: &'static str, _variant_index: u32, _variant: &'static str) -> Result<Self::Ok, Self::Error> {
            $get_result("unit_variant")
        }
    };

    ($get_result:expr, newtype_struct) => {
        fn serialize_newtype_struct<T: ?Sized + serde::ser::Serialize>(self, _name: &'static str, _value: &T) -> Result<Self::Ok, Self::Error> {
            $get_result("newtype_struct")
        }
    };

    ($get_result:expr, newtype_variant) => {
        fn serialize_newtype_variant<T: ?Sized + serde::ser::Serialize>(self, _name: &'static str, _variant_index: u32, _variant: &'static str, _value: &T) -> Result<Self::Ok, Self::Error> {
            $get_result("newtype_variant")
        }
    };

    ($get_result:expr, seq) => {
        fn serialize_seq(self, _len: Option<usize>) -> Result<Self::SerializeSeq, Self::Error> {
            $get_result("seq")
        }
    };

    ($get_result:expr, tuple) => {
        fn serialize_tuple(self, _len: usize) -> Result<Self::SerializeTuple, Self::Error> {
            $get_result("tuple")
        }
    };

    ($get_result:expr, tuple_struct) => {
        fn serialize_tuple_struct(self, _name: &'static str, _len: usize) -> Result<Self::SerializeTupleStruct, Self::Error> {
            $get_result("tuple_struct")
        }
    };

    ($get_result:expr, tuple_variant) => {
        fn serialize_tuple_variant(self, _name: &'static str, _variant_index: u32, _variant: &'static str, _len: usize) -> Result<Self::SerializeTupleVariant, Self::Error> {
            $get_result("tuple_variant")
        }
    };

    ($get_result:expr, map) => {
        fn serialize_map(self, _len: Option<usize>) -> Result<Self::SerializeMap, Self::Error> {
            $get_result("map")
        }
    };

    ($get_result:expr, struct) => {
        fn serialize_struct(self, _name: &'static str, _len: usize) -> Result<Self::SerializeStruct, Self::Error> {
            $get_result("struct")
        }
    };

    ($get_result:expr, struct_variant) => {
        fn serialize_struct_variant(self, _name: &'static str, _variant_index: u32, _variant: &'static str, _len: usize) -> Result<Self::SerializeStructVariant, Self::Error> {
            $get_result("struct_variant")
        }
    };
}

#[cfg(test)]
mod tests {
    use serde::ser;
    use thiserror::Error;

    #[derive(Debug, Error)]
    enum SerializationError {
        #[error("Cannot serialize value from {0}")]
        CannotSerializeFrom(String),
        #[error("Custom({0})")]
        Custom(String)
    }

    impl serde::ser::Error for SerializationError {
        fn custom<T>(msg:T) -> Self
        where T: std::fmt::Display
        {
            SerializationError::Custom(msg.to_string())
        }
    }

    #[test]
    fn all_get_results() {
        struct MySerializer;

        impl ser::Serializer for MySerializer {
            type Error = SerializationError;
            type Ok = ();

            type SerializeMap = ser::Impossible<Self::Ok, Self::Error>;
            type SerializeSeq = ser::Impossible<Self::Ok, Self::Error>;
            type SerializeStruct = ser::Impossible<Self::Ok, Self::Error>;
            type SerializeStructVariant = ser::Impossible<Self::Ok, Self::Error>;
            type SerializeTuple = ser::Impossible<Self::Ok, Self::Error>;
            type SerializeTupleStruct = ser::Impossible<Self::Ok, Self::Error>;
            type SerializeTupleVariant = ser::Impossible<Self::Ok, Self::Error>;
            
            impl_serialize!(|value_type: &str| Err(SerializationError::CannotSerializeFrom(value_type.to_string())), [
                bool,
                bytes,
                i8, i16, i32, i64,
                u8, u16, u32, u64,
                f32, f64,
                char,
                str,
                none, some, unit,
                unit_struct, unit_variant,
                newtype_struct, newtype_variant,
                seq, map,
                tuple, tuple_struct, tuple_variant,
                struct, struct_variant
            ]);
        }
    }
}